--[[
	r5-beta
NEW FEATURES:

--- User Interface ---
• 
	r4-beta
FIXES:

--- Elemental ---
• An error will no longer occur when using Bloodlust.
• The progress bars for the 3 elementals will now be their correct size.
• The 3 lightning orbs will now always appear when using Stormseeker.

--- Restoration ---
• Lightning Surge Totem's icon now properly appears
• Earthen Shield Totem's health/progress bar now correctly appears
• An Earthen Shield Totem icon has been added to the "left" column of small icons.
• All icons will become opaque when in combat, or targeting a friendly target.
• The Tidal Waves bar will now appear and disappear appropriately.
• Spiritwalker's Grace's duration timer bar now correctly appears.

--- User Interface ---
• The interface option category should now appear in the navigation panel properly.

]]
local FOLDER_NAME, SSA = ...

local Auras = LibStub("AceAddon-3.0"):NewAddon("ShamanAuras", "AceConsole-3.0","AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0");
Auras.version = GetAddOnMetadata(FOLDER_NAME, "Version")
Auras:RegisterChatCommand("ssa","ChatCommand");
local L = LibStub("AceLocale-3.0"):GetLocale("ShamanAuras", true)

--local obj = {}
local caster,cooldown,duration = {},{},{};
local smallSpells = {};
local spellobj,totemObj,targetObj,miscObj,spells,ignoreSpells,GUIDs = {},{},{},{},{},{},{};
local loadedBuffBars,loadedMainBars,loadedUtilBars = {},{},{};
local talents,execTime = {},{};
local isFireStormElemental,isEarthElemental,isTotemMastery = false,false,false;
local spec = nil;
-- Booleans
local isIcefury = false;
local tempBool = false;
-------------------------------------------------------------------------------------------------------
----- Initialize Global Variables
-------------------------------------------------------------------------------------------------------
SSA.BackdropSB = {
	bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,
	tileSize = 16,
	edgeSize = 10
}
SSA.ButtonFont = CreateFont("SSAButtonFont");
local ButtonFont = SSA.ButtonFont;
ButtonFont:SetFont("Interface\\addons\\ShamanAuras\\Media\\fonts\\Continuum_Medium.ttf",12,"OUTLINE");
ButtonFont:SetJustifyH("CENTER");


SSA.MoveEnh = CreateFrame("Frame","MoveEnh",UIParent);
SSA.MoveRes = CreateFrame("Frame","MoveRes",UIParent);

------------------------------
--- For Debugging Purposes ---
------------------------------
SSA.ErrorFrame = CreateFrame("Frame",nil,UIParent);
_G["SSA_ErrorFrame"] = SSA.ErrorFrame;
local ErrorFrame = SSA.ErrorFrame;
ErrorFrame:SetWidth(800);
ErrorFrame:SetHeight(180);
ErrorFrame:SetPoint("CENTER",UIParent,"CENTER",0,250);
ErrorFrame:Show();

ErrorFrame.text = ErrorFrame:CreateFontString(nil, "MEDIUM", "GameFontHighlightLarge");
ErrorFrame.text:SetPoint("CENTER",0,-200);
ErrorFrame.text:SetFont("Interface\\addons\\ShamanAuras\\Media\\fonts\\Continuum_Medium.ttf", 10,"OUTLINE");
ErrorFrame.text:SetTextColor(1,0.75,0,1);
ErrorFrame.text:SetJustifyH("LEFT");

function Auras:CurText()
	return SSA.ErrorFrame.text:GetText() or '';
end

local Frame = CreateFrame("Frame");
		
-- Frame Tables
local CDFrame = {
	[1] = {
	},
	[2] = {
	},
	[3] = {
	}
}
SSA.isEoE = false;
SSA.spec = nil;
local spec = nil;
SSA.ChargeCDFrame = {
	[1] = {
	},
	[2] = {
	},
	[3] = {
	}
}



-------------------------------------------------------------------------------------------------------
----- Local Functions
-------------------------------------------------------------------------------------------------------

local function AdjustIconDisplay(object,r,g,b,a)
	if (SSA.spec ~= 2) then
		object.texture:SetVertexColor(r,g,b,a);
		object.CD.text:SetAlpha(a);
		if (SSA[object:GetName().."Glow"]) then
			SSA[object:GetName()]:SetAlpha(a);
		end
		if (object.Charges) then
			object.Charges.text:SetAlpha(a);
		end
	end
end

local function AlterTexture(texture,bool,alpha)
	texture:SetDesaturated(bool);
	texture:SetAlpha(alpha);
end

local function GetMaelstromBarObj(curSpec)
	if (curSpec == 1) then
		return SSA.MaelstromBarEle,Auras.db.char.triggers.ele.maelstrom;
	elseif (curSpec == 2) then
		--return SSA.MaelstromBarEnh,Auras.db.char.triggers.enh.maelstrom;
	end
end

-- Retrieve object by SpellID
local function GetObject(spellID)
	if (spellObj[spec][spellID]) then
		return spellObj[spec][spellID];
	elseif (totemObj[spec][spellID]) then
		return totemObj[spec][spellID]
	elseif (totemObj[0][spellID]) then
		return totemObj[0][spellID];
	else
		return nil;
	end
end

-- Retrieve Specialization Name
local function GetSpecName(spec)
	if (spec == 1) then
		return "Ele";
	elseif (spec == 2) then
		return "Enh";
	elseif (spec == 3) then
		return "Res";
	end
end

-- Retrieve Spell Info Name
local function GetSpellName(spellID)
	local name = GetSpellInfo(spellID);
	return name;
end

-- Return the type of player power being used
local function GetPowerInfo(powerType)
	if (powerType == "Maelstrom") then
		local playerPower,maxPower = UnitPower('player',11),UnitPowerMax('player',11);
		
		return playerPower,maxPower;
	elseif (powerType == "Icefury") then
		--local name = GetSpellInfo(210714);
		local _,_,_,count = UnitBuff('player',GetSpellName(210714));

		return count or 0;
	end
end

-- Aura Group Builder
local function InitializeGroup(group,width,height,point,relativeTo,relativePoint,x,y,spec)
	--local spec = SSA.spec;
	if (width ~= nil) then
		group:SetWidth(width);
		group:SetHeight(height);
	end

	group:EnableMouse(false);
	group:SetMovable(false);
	group:RegisterForDrag("LeftButton");
	group:SetPoint(point,(SSA[relativeTo] or UIParent),relativePoint,x,y)
	if (not Auras.db.char.aura[spec][group:GetName()]) then
		--SSA.ErrorFrame.text:SetText(Auras:CurText().."[X]: "..group:GetName().."\n");
		group:Hide();
	--else
		--group:Hide();
	else
		--SSA.ErrorFrame.text:SetText(Auras:CurText().."[ ]: "..group:GetName().."\n");
	end
end

-- Initialize min/max values of a status bar
local function InitializeStatusBar(statusbar,minVal,maxVal,curVal,textVal)
	if (statusbar) then
		statusbar:SetMinMaxValues(minVal,maxVal);
		statusbar:SetValue(curVal or 0);

		if (textVal) then
			statusbar.text:SetText(textVal);
		end
	end
end

-- Check if the current target is an enemy
local function IsEnemyTarget()
	if (UnitExists('target') and not UnitIsFriend('target','player') and not UnitIsDead('target')) then
		return true;
	else
		return false;
	end
end

-- Set the Maelstrom Bar's alpha based on whether the target is an enemy and if the player is in combat
local function MaelstromAlpha()
	if (SSA.spec == 1) then
		local MaelstromBar = GetMaelstromBarObj(SSA.spec);
		if (IsEnemyTarget()) then
			--MaelstromBar:Show();
			MaelstromBar:SetAlpha(1);
			MaelstromBar.Elapsed = 0;
			if (not UnitAffectingCombat('player')) then
				MaelstromBar:SetAlpha(0.5);
			else
				MaelstromBar:SetAlpha(1);
			end
		else
			if (not UnitAffectingCombat('player')) then
				MaelstromBar:SetAlpha(0);
			end
		end
	end
end

-- Convert time, in seconds, to a timer string
local function parseTime(timer,precision)
	if (timer ~= nil and timer > 0) then
		local m,s,ms,fs = math.floor(timer / 60),math.floor(math.fmod(timer,60)),(("%%.%df"):format(1)):format(math.fmod(timer,60),1),timer;

		if (m >= 1 and s == 0) then
			return m..":00",fs;
		else
			if (m > 0) then
				m = m..":";
				if (s < 10) then
					s = "0"..s;
				end
			else
				m = '';
				if (s < 0) then
					s = '';
				end
			end
		end
		if (precision and s < 60) then
			return m..ms,fs;
		else
			return m..s,fs;
		end
	else
		return 0,0;
	end
end

local function ToggleCooldownSwipe(object,arg1,arg2)
	if (not arg1) then
		_G[object:GetName()]:SetSwipeColor(1,1,1,0);
		_G[object:GetName()]:SetDrawSwipe(arg1)
		_G[object:GetName()]:SetDrawEdge(arg1)
		_G[object:GetName()]:SetDrawBling(arg1)
	else
		_G[object:GetName()]:SetSwipeColor(1,1,1,1);
	end
	--_G[object:GetName()]:SetAlpha(0);
	--[[_G[object:GetName()]:SetDrawBling(arg1)
	-- Hides background (swipe texture) of the cooldown
	_G[object:GetName()]:SetDrawSwipe(arg1)
	-- Hides the glowing line that marks the edge of the cooldown swipe
	_G[object:GetName()]:SetDrawEdge(arg1)
	-- Hides the countdown numbers regardless of the cvar setting]]
	_G[object:GetName()]:SetHideCountdownNumbers(arg2)
	--SSA.ErrorFrame.text:SetText("Name: "..object:GetObjectType().."\n");
end

-- Run the cooldown for any abilities that are currently on cooldown
local function ExecuteCooldown()
	if (SSA.spec ~= 2) then
		local fontSize = 0;
		local spec = SSA.spec;

		for spellID,startTime in pairs(Auras.db.char.cooldowns) do
			--local name = GetSpellInfo(spellID);
			local _,duration = GetSpellCooldown(GetSpellName(spellID));
			
			if ((SSA.isEoE and (SSA.chargedSpells[SSA.spec][spellID] or SSA.chargedSpells[SSA.spec][spellID])) or spellID == 201897) then
				-- Do Nothing
			else
				if ((duration or 0) > 2 and spells[spec][spellID]) then
					local CDFrame = SSA[spellObj[spec][spellID]:GetName()].CD;

					CDFrame:SetCooldown(startTime,duration);
					CDFrame:Show()
					ToggleCooldownSwipe(CDFrame,true,false);
					--CDFrame:SetSwipeColor(1,1,1,1);
					CDFrame:SetScript("OnUpdate", function(self)
						--local charges,maxcharges = SpellCharges(spellID);
						local PlayerGUID = UnitGUID('player');
						local expires = startTime + duration;
						local timer = expires - GetTime();
						local remaining,seconds = parseTime(timer,false);
						
						if (smallSpells[spellID]) then
							if (seconds > 60) then
								fontSize = 11;
							else
								fontSize = 16;
							end
						else
							if (seconds > 60) then
								fontSize = 14;
							else
								fontSize = 18;
							end
						end
						
						if (spellID ~= 188838) then
							if ((seconds or 0) > 0.1) then
								self.text:SetFont("Interface\\addons\\ShamanAuras\\media\\fonts\\PT_Sans_Narrow.TTF", fontSize,"OUTLINE");
								self.text:SetText(parseTime(timer,false));
							else
								Auras.db.char.cooldowns[spellID] = nil;
								self.text:SetText('');
								self:SetScript("OnUpdate",nil);
							end
						end
					end);
				end
			end
		end
	end
end

-- Return the number of charges an ability currently has by spellID
local function SpellCharges(spellid)
	local charges,maxcharges = GetSpellCharges(spellid);
	local isMaxCharge = false;
	
	if (charges == maxcharges) then
		isMaxCharge = true;
	end
	
	if ((charges or 0) > 0) then
		return charges,isMaxCharge;
	else
		return nil,false;
	end
end

local function TotemMasteryCheck()
	local spec = SSA.spec;
	local isEle = false;
	local isLearned = select(4,GetTalentInfo(1,3,1));
	local isTotem = false;
	local isRefresh = false;
	local isBuff = false; 
	local buff = UnitBuff('player',GetSpellName(202192));
	
	if (not buff) then
		isBuff = true;
	end
	local remaining
	if (spec == 1) then
		isEle = true;
	end
	for i=1,5 do
		if (select(2,GetTotemInfo(i)) == "Totem Mastery") then
			isTotem = true;
			remaining = (select(3,GetTotemInfo(i)) + select(4,GetTotemInfo(i))) - GetTime();
			break;
		end
	end
	
	if (remaining and remaining <= Auras.db.char.triggers.ele.totemMastery) then
		isRefresh = true;
	end
	
	--[[SSA.ErrorFrame.text:SetText("isEle: "..tostring(isEle).."\n");
	SSA.ErrorFrame.text:SetText(Auras:CurText().."isLearned: "..tostring(isLearned).."\n");
	SSA.ErrorFrame.text:SetText(Auras:CurText().."isTotem: "..tostring(isTotem).."\n");
	SSA.ErrorFrame.text:SetText(Auras:CurText().."isRefresh: "..tostring(isRefresh).."\n");
	SSA.ErrorFrame.text:SetText(Auras:CurText().."isBuff: "..tostring(buff).."\n");
	SSA.ErrorFrame.text:SetText(Auras:CurText().."isEnabled: "..tostring(Auras.db.char.aura[1].TotemMastery).."\n");
	SSA.ErrorFrame.text:SetText(Auras:CurText().."Remaining: "..tostring(math.floor(remaining or 0)).."\n");]]
	if ((not isTotem or isRefresh or not buff) and isEle and isLearned and Auras.db.char.aura[1].TotemMastery) then
		--SSA.ErrorFrame.text:SetText(Auras:CurText().."Response: true");
		return true;
	--elseif ((isTotem or isRefresh or buff) and isEle and isLearned and Auras.db.char.aura[1].TotemMastery) then
	else
		--SSA.ErrorFrame.text:SetText(Auras:CurText().."Response: false");
		return false;
	end

	
	--return isEle and isLearned and isNoTotem and isNoRefresh and isBuff and Auras.db.char.aura[1].TotemMastery;
end

-- Determine target type and appropriate set icon alpha and color
local function TargetCheck()
	
end

-------------------------------------------------------------------------------------------------------
----- Global Functions
-------------------------------------------------------------------------------------------------------

-- Show/Hide aura icons based on the current spec
function Auras:ToggleAuraIcons(curSpec)
	--if (curSpec ~= 2) then
		for i=1,3 do
			for spellID in pairs(spellObj[i]) do
				local name = spellObj[i][spellID]:GetName();

				if (spellObj[i][spellID]) then
					if (i == curSpec and (Auras.db.char.aura[i][name]) and curSpec ~= 2) then
						if (SSA[name.."Glow"]) then
							SSA[name.."Glow"]:Show();
						end
						
						if (SSA[name].Charges) then
							SSA[name].Charges:Show();
						end
						
						spellObj[i][spellID]:Show();
					else
						if (SSA[name.."Glow"]) then
							SSA[name.."Glow"]:Hide();
						end
						if (SSA[name.."Charges"]) then
							SSA[name.."Charges"]:Hide();
						end
						spellObj[i][spellID]:Hide();
					end
				end
			end
			for j=1,getn(miscObj[i]) do
				if (i == curSpec and curSpec ~= 2) then
					if (miscObj[i][j]:GetName() == "MaelstromBarEnh" or miscObj[i][j]:GetName() == "MaelstromBarEle") then
						MaelstromAlpha();
					else
						miscObj[i][j]:Show();
					end
				else
					miscObj[i][j]:Hide();
				end
			end
		end
	--end
end

-- Event: ADDON_LOADED
function Auras:OnInitialize()
	local defaults = {
		char = {
			isEleFirst = true,
			isEnhFirst = true,
			isResFirst = true,
			cooldowns = {},
			info = {
				totems = {
					eShield = {
					},
					cBurst = {
						absorb = 0;
					}
				}
			},
			triggers = {
				ele = {
					MaelstromAnim = true,
					maelstrom = 90,
					flameShock = 10,
					totemMastery = 15,
					OoCAlpha = 0.5,
					OoRColor = {
						r = 1,
						g = 0,
						b = 0,
						a = 1,
					}
				},
				enh = {
					OoCAlpha = 0.5,
					OoRColor = {
						r = 1,
						g = 0,
						b = 0,
						a = 1,
					}
				},
				res = {
					OoCAlpha = 0.5,
				},
				default = {
					ele = {
						MaelstromAnim = true,
						maelstrom = 90,
						flameShock = 10,
						totemMastery = 15,
						OoCAlpha = 0.5,
						OoRColor = {
							r = 1,
							g = 0,
							b = 0,
							a = 1,
						}
					},
					enh = {
						OoCAlpha = 0.5,
						OoRColor = {
							r = 1,
							g = 0,
							b = 0,
							a = 1,
						}
					},
					res = {
						OoCAlpha = 0.5,
					},
				},
			},
			aura = {
				--minor = {
					[1] = {
						AncestralGuidanceEle = true,
						AncestralGuidanceBarEle = true,
						AscendanceEle = true,
						AscendanceBarEle = true,
						AstralShiftEle = true,
						AstralShiftBarEle = true,
						AuraGroupEle = true,
						Bloodlust = true,
						BloodlustBarEle = true,
						BuffTimerBarGrpEle = true,
						CleanseSpiritEle = true,
						EarthElemental = true,
						EarthElementalBar = true,
						EarthgrabTotemEle = true,
						EarthgrabTotemBarEle = true,
						EarthShock = true,
						EarthShockGlow = true,
						ElementalBlast = true,
						ElementalBlastCritBar = true,
						ElementalBlastHasteBar = true,
						ElementalBlastMasteryBar = true,
						ElementalMastery = true,
						ElementalMasteryBar = true,
						FireElemental = true,
						FireElementalBar = true,
						FlameShock = true,
						FlameShockGlow = true,
						FrostShock = true,
						GustWindEle = true,
						HexEle = true,
						HexBarEle = true,
						Icefury = true,
						IcefuryBar = true,
						LargeIconGrpBotEle = true,
						LargeIconGrpTopEle = true,
						LavaBurstEle = true,
						LavaBurstEleGlow = true,
						LavaBurstEleCharges = true,
						LavaSurgeEle = true,
						--LavaSurgeBar = true,
						LightningSurgeTotemEle = true,
						LightningSurgeTotemBarEle = true,
						LiquidMagmaTotem = true,
						LiquidMagmaTotemBar = true,
						MaelstromBarEle = true,
						MainTimerBarGrpEle = true,
						SmallIconGrpLeftEle = true,
						SmallIconGrpRightEle = true,
						StormElemental = true,
						StormElementalBar = true,
						Stormkeeper = true,
						StormkeeperChargeGrp = true,
						Thunderstorm = true,
						TotemMastery = true,
						TotemMasteryTexture = true,
						UtilTimerBarGrpEle = true,
						VoodooTotemEle = true,
						VoodooTotemBarEle = true,
						WindRushTotemEle = true,
						WindRushTotemBarEle = true,
						WindShearEle = true
					},
					[2] = {
						AscendanceEnh = true,
						AscendanceBarEnh = true,
						AstralShiftEnh = true,
						AstralShiftBarEnh = true,
						AuraGroupEnh = true,
						BloodlustEnh = true,
						BloodlustBarEnh = true,
						Boulderfist = true,
						BoulderfistBar = true,
						BoulderfistGlow = true,
						BoulderfistCharges = true,
						BuffTimerBarGrpEnh = true,
						CleanseSpiritEnh = true,
						CrashLightning = true,
						CrashLightningBar = true,
						CrashLightningGlow = true,
						EarthenSpike = true,
						EarthenSpikeBar = true,
						EarthgrabTotemEnh = true,
						EarthgrabTotemBarEnh = true,
						FeralLunge = true,
						FeralSpirit = true,
						FeralSpiritBar = true,
						Flametongue = true,
						FlametongueBar = true,
						FlametongueGlow = true,
						Frostbrand = true,
						FrostbrandBar = true,
						FrostbrandGlow = true,
						HexEnh = true,
						HexBarEnh = true,
						LargeIconGrpBotEnh = true,
						LargeIconGrpTopEnh = true,
						LightningSurgeTotemEnh = true,
						LightningSurgeTotemBarEnh = true,
						MaelstromBarEnh = true,
						MainTimerBarGrpEnh = true,
						Rainfall = true,
						SmallIconGrpLeftEnh = true,
						SmallIconGrpRightEnh = true,
						SpiritWalk = true,
						SpiritWalkBar = true,
						Stormstrike = true,
						StormstrikeChargeGrp = true,
						StormstrikeChargesText = false,
						StormstrikeGlow = true,
						Sundering = true,
						UtilTimerBarGrpEnh = true,
						VoodooTotemEnh = true,
						VoodooTotemBarEnh = true,
						WindRushTotemEnh = true,
						WindRushTotemBarEnh = true,
						WindShearEnh = true,
						Windsong = true,
						WindsongBar = true,
						WindsongGlow = true,
					},
					[3] = {
						AncestralGuidanceRes = true,
						AncestralGuidanceBarRes = true,
						AncestralProtectionTotem = true,
						AncestralProtectionTotemBar = true,
						AscendanceRes = true,
						AscendanceBarRes = true,
						AstralShiftRes = true,
						AstralShiftBarRes = true,
						AuraGroupRes = true,
						--Bloodlust = true,
						BloodlustBarRes = true,
						BuffTimerBarGrpRes = true,
						CloudburstTotem = true,
						CloudburstTotemBar = true,
						EarthenShieldTotem = true,
						EarthenShieldTotemBar = true,
						EarthgrabTotemRes = true,
						EarthgrabTotemBarRes = true,
						FlameShockRes = true,
						FlameshockResGlow = true,
						GiftOfQueen = true,
						GustWindRes = true,
						HealingRain = true,
						HealingStreamTotem = true,
						HealingStreamTotemBar = true,
						HealingStreamCharges = true,
						HealingTideTotem = true,
						HealingTideTotemBar = true,
						HexRes = true,
						HexBarRes = true,
						LargeIconGrpBotRes = true,
						LargeIconGrpTopRes = true,
						LavaBurstRes = true,
						LavaBurstResGlow = true,
						LavaBurstResCharges = true,
						LavaSurgeRes = true,
						--LavaSurgeBarRes = true,
						LightningSurgeTotemRes = true,
						LightningSurgeTotemBarRes = true,
						MainTimerBarGrpRes = true,
						PurifySpirit = true,
						Riptide = true,
						RiptideCharge = true,
						SmallIconGrpLeftRes = true,
						SmallIconGrpRightRes = true,
						SpiritLinkTotem = true,
						SpiritLinkTotemBar = true,
						SpiritwalkersGrace = true,
						SpiritwalkersGraceBar = true,
						TidalWaves = true,
						TidalWavesBar = true,
						TidalWavesCharges = true,
						UnleashLife = true,
						UnleashLifeBar = true,
						UtilTimerBarGrpRes = true,
						VoodooTotemRes = true,
						VoodooTotemBarRes = true,
						Wellspring = true,
						WindRushTotemRes = true,
						WindRushTotemBarRes = true,
						WindShearRes = true,
					},
				--},
			},
			frames = {
				timerbars = {
					buff = {
						[1] = true,
						[2] = true,
						[3] = true,
					},
					main = {
						[1] = true,
						[2] = true,
						[3] = true,
					},
					util = {
						[1] = true,
						[2] = true,
						[3] = true,
					},
				},
				eleGrp = {
					AuraGroupEle = {
						point = "CENTER",
						relativeTo = nil,
						relativePoint = "CENTER",
						x = 0,
						y = 0,
						width = (GetScreenWidth() - 500),
						height = (GetScreenHeight() - 250),
					},
					LargeIconGrpTopEle = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 0,
						y = -180,
						width = 250,
						height = 50,
					},
					LargeIconGrpBotEle = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 0,
						y = -225,
						width = 250,
						height = 50,
					},
					SmallIconGrpLeftEle = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = -175,
						y = -170,
						width = 35,
						height = 160,
					},
					SmallIconGrpRightEle = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 175,
						y = -170,
						width = 35,
						height = 160,
					},
					BuffTimerBarGrpEle = {
						point = "LEFT",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 190,
						y = -125,
						width = 115,
						height = 180,
					},
					MainTimerBarGrpEle = {
						point = "RIGHT",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = -190,
						y = -125,
						width = 90,
						height = 180,
					},
					UtilTimerBarGrpEle = {
						point = "RIGHT",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = -190,
						y = -125,
						width = 70,
						height = 180,
					},
					MaelstromBarEle = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 0,
						y = -139,
						width = nil,
						height = nil,
						alpha = 0.5,
					},
					IcefuryBar = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 0,
						y = -110,
						width = nil,
						height = nil,
						alpha = 0.9,
					},
					TotemMastery = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 0,
						y = -210,
						width = nil,
						height = nil,
					},
					LavaSurgeEle = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = -130,
						y = 0,
						width = nil,
						height = nil,
					},
					StormkeeperChargeGrp = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = 0,
						y = -110,
						width = 260,
						height = 50,
					}
				},
				enhGrp = {
					AuraGroupEnh = {
						point = "CENTER",
						relativeTo = nil,
						relativePoint = "CENTER",
						x = 0,
						y = 0,
						width = (GetScreenWidth() - 500),
						height = (GetScreenHeight() - 250),
					},
					LargeIconGrpTopEnh = {
						point = "CENTER",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = 0,
						y = -180,
						width = 250,
						height = 50,
					},
					LargeIconGrpBotEnh = {
						point = "CENTER",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = 0,
						y = -225,
						width = 250,
						height = 50,
					},
					SmallIconGrpLeftEnh = {
						point = "CENTER",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = -175,
						y = -170,
						width = 35,
						height = 160,
					},
					SmallIconGrpRightEnh = {
						point = "CENTER",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = 175,
						y = -170,
						width = 35,
						height = 160,
					},
					BuffTimerBarGrpEnh = {
						point = "LEFT",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = 190,
						y = -125,
						width = 131,
						height = 180,
					},
					MainTimerBarGrpEnh = {
						point = "RIGHT",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = -190,
						y = -125,
						width = 131,
						height = 180,
					},
					UtilTimerBarGrpEnh = {
						point = "RIGHT",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = -190,
						y = -125,
						width = 47,
						height = 180,
					},
					MaelstromBarEnh = {
						point = "CENTER",
						relativeTo = "AuraGroupEnh",
						relativePoint = "CENTER",
						x = 0,
						y = -139,
						width = nil,
						height = nil,
					},
				},
				resGrp = {
					AuraGroupRes = {
						point = "CENTER",
						relativeTo = nil,
						relativePoint = "CENTER",
						x = 0,
						y = 0,
						width = (GetScreenWidth() - 500),
						height = (GetScreenHeight() - 250),
					},
					LargeIconGrpTopRes = {
						point = "CENTER",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = 0,
						y = -180,
						width = 250,
						height = 50,
					},
					LargeIconGrpBotRes = {
						point = "CENTER",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = 0,
						y = -225,
						width = 250,
						height = 50,
					},
					SmallIconGrpLeftRes = {
						point = "CENTER",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = -175,
						y = -170,
						width = 35,
						height = 160,
					},
					SmallIconGrpRightRes = {
						point = "CENTER",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = 175,
						y = -170,
						width = 35,
						height = 160,
					},
					BuffTimerBarGrpRes = {
						point = "LEFT",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = 190,
						y = -125,
						width = 131,
						height = 180,
					},
					MainTimerBarGrpRes = {
						point = "RIGHT",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = -190,
						y = -125,
						width = 105,
						height = 180,
					},
					UtilTimerBarGrpRes = {
						point = "RIGHT",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = -190,
						y = -125,
						width = 47,
						height = 180,
					},
					EarthenShieldTotemBar = {
						point = "CENTER",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = 0,
						y = -139,
						width = nil,
						height = nil,
					},
					TidalWavesBar = {
						point = "CENTER",
						relativeTo = "AuraGroupRes",
						relativePoint = "CENTER",
						x = 0,
						y = -202,
						width = nil,
						height = nil,
					},
					LavaSurgeRes = {
						point = "CENTER",
						relativeTo = "AuraGroupEle",
						relativePoint = "CENTER",
						x = -130,
						y = 0,
						width = nil,
						height = nil,
					},
				},
				defaultPos = {
					eleGrp = {
						AuraGroupEle = {
							point = "CENTER",
							relativeTo = nil,
							relativePoint = "CENTER",
							x = 0,
							y = 0,
						},
						LargeIconGrpTopEle = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 0,
							y = -180,
						},
						LargeIconGrpBotEle = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 0,
							y = -225,
						},
						SmallIconGrpLeftEle = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = -175,
							y = -170,
						},
						SmallIconGrpRightEle = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 175,
							y = -170,
						},
						BuffTimerBarGrpEle = {
							point = "LEFT",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 190,
							y = -125,
						},
						MainTimerBarGrpEle = {
							point = "RIGHT",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = -190,
							y = -125,
						},
						UtilTimerBarGrpEle = {
							point = "RIGHT",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = -190,
							y = -125,
						},
						MaelstromBarEle = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 0,
							y = -139,
						},
						IcefuryBar = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 0,
							y = -110,
						},
						TotemMastery = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 0,
							y = -210,
						},
						LavaSurgeEle = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = -130,
							y = 0,
						},
						StormkeeperChargeGrp = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = 0,
							y = -110,
						}
					},
					enhGrp = {
						AuraGroupEnh = {
							point = "CENTER",
							relativeTo = nil,
							relativePoint = "CENTER",
							x = 0,
							y = 0,
						},
						LargeIconGrpTopEnh = {
							point = "CENTER",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = 0,
							y = -180,
						},
						LargeIconGrpBotEnh = {
							point = "CENTER",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = 0,
							y = -225,
						},
						SmallIconGrpLeftEnh = {
							point = "CENTER",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = -175,
							y = -170,
						},
						SmallIconGrpRightEnh = {
							point = "CENTER",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = 175,
							y = -170,
						},
						BuffTimerBarGrpEnh = {
							point = "LEFT",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = 190,
							y = -125,
						},
						MainTimerBarGrpEnh = {
							point = "RIGHT",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = -190,
							y = -125,
						},
						UtilTimerBarGrpEnh = {
							point = "RIGHT",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = -190,
							y = -125,
						},
						MaelstromBarEnh = {
							point = "CENTER",
							relativeTo = "AuraGroupEnh",
							relativePoint = "CENTER",
							x = 0,
							y = -139,
						},
					},
					resGrp = {
						AuraGroupRes = {
							point = "CENTER",
							relativeTo = nil,
							relativePoint = "CENTER",
							x = 0,
							y = 0,
						},
						LargeIconGrpTopRes = {
							point = "CENTER",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = 0,
							y = -180,
						},
						LargeIconGrpBotRes = {
							point = "CENTER",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = 0,
							y = -225,
						},
						SmallIconGrpLeftRes = {
							point = "CENTER",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = -175,
							y = -170,
						},
						SmallIconGrpRightRes = {
							point = "CENTER",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = 175,
							y = -170,
						},
						BuffTimerBarGrpRes = {
							point = "LEFT",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = 190,
							y = -125,
						},
						MainTimerBarGrpRes = {
							point = "RIGHT",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = -190,
							y = -125,
						},
						UtilTimerBarGrpRes = {
							point = "RIGHT",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = -190,
							y = -125,
						},
						EarthenShieldTotemBar = {
							point = "CENTER",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = 0,
							y = -139,
						},
						TidalWavesBar = {
							point = "CENTER",
							relativeTo = "AuraGroupRes",
							relativePoint = "CENTER",
							x = 0,
							y = -202,
						},
						LavaSurgeRes = {
							point = "CENTER",
							relativeTo = "AuraGroupEle",
							relativePoint = "CENTER",
							x = -130,
							y = 0,
						},
					},
				},
			},
		}
	}
	local about_panel = LibStub:GetLibrary("LibAboutPanel", true)

	if about_panel then
		self.optionsFrame = about_panel.new(nil, "ShamanAuras")
	end
	
	self.db = LibStub("AceDB-3.0"):New("SSA_db",defaults);
	self:SetupOptions();
	
	
end

-- Event: PLAYER_LOGIN
function Auras:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN");
	self:RegisterEvent("ADDON_LOADED");
	self:RegisterEvent("PLAYER_LOGIN");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("PLAYER_LEVEL_UP");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("UNIT_POWER");
	self:RegisterEvent("PLAYER_TARGET_CHANGED");
	self:RegisterEvent("PLAYER_TALENT_UPDATE");
	self:RegisterEvent("SPELL_UPDATE_CHARGES","ChargeCooldown");

	smallSpells = {
		[108281] = true, -- Ancestral Guidance
		[108271] = true, -- Astral Shift
		[51886]  = true, -- Cleanse Spirit
		[198103] = true, -- Earth Elemental
		[51485]  = true, -- Earthgrab Totem
		[192063] = true, -- Gust of Wind
		[51514]  = true, -- Hex (Frog)
		[210873] = true, -- Hex (Raptor)
		[211015] = true, -- Hex (Roach)
		[211010] = true, -- Hex (Snake)
		[211004] = true, -- Hex (Spider)
		[192058] = true, -- Lightning Surge Totem
		[77130]  = true, -- Purify Spirit
		[215864] = true, -- Rainfall
		[58875]  = true, -- Spirit Walk
		[79206]  = true, -- Spiritwalker's Grace
		[51490]  = true, -- Thunderstorm
		[196932] = true, -- Voodoo Totem
		[192077] = true, -- Wind Rush Totem
		[57994]  = true, -- Wind Shear
	}
	
	spells = {
		[1] = { -- Elemental
			[108281] = true, -- Ancestral Guidance
			[114050] = true, -- Ascendance
			[108271] = true, -- Astral Shift
			--[2825]   = true, -- Bloodlust
			[51886]  = true, -- Cleanse Spirit
			[117014] = true, -- Elemental Blast
			[198103] = true, -- Earth Elemental
			[188616] = true, -- Earth Elemental (SPELL_SUMMON ID)
			[51485]  = true, -- Earthgrab Totem
			[8042]   = true, -- Earth Shock
			[16166]  = true, -- Elemental Mastery
			[198067] = true, -- Fire Elemental
			[188592] = true, -- Fire Elemental (SPELL_SUMMON ID)
			[188389] = true, -- Flame Shock
			[196840] = true, -- Frost Shock
			[192063] = true, -- Gust of Wind
			[51514]  = true, -- Frog Hex
			[210873] = true, -- Raptor Hex
			[211004] = true, -- Spider Hex
			[211010] = true, -- Snake Hex
			[211015] = true, -- Roach Hex
			[210714] = true, -- Icefury
			[51505]  = true, -- Lava Burst
			[77762]  = true, -- Lava Surge
			[192058] = true, -- Lightning Surge Totem
			[192222] = true, -- Liquid Magma Totem
			[192249] = true, -- Storm Elemental
			[157299] = true, -- Storm Elemental (SPELL_SUMMON ID)
			[205495] = true, -- Stormkeeper
			[51490]  = true, -- Thunderstorm
			[196932] = true, -- Voodoo Totem
			[192077] = true, -- Wind Rush Totem
			[57994]  = true, -- Wind Shear
		},
		[2] = { -- Enhancement
			[114051] = true, -- Ascendance
			[108271] = true, -- Astral Shift
			--[2825]   = true, -- Bloodlust
			[201897] = true, -- Boulderfist
			[218825] = true, -- Boulderfist (SPELL_AURA_APPLIED ID)
			[51886]  = true, -- Cleanse Spirit
			[187874] = true, -- Crash Lightning
			[187878] = true, -- Crash Lightning (SPELL_AURA_APPLIED ID)
			[188089] = true, -- Earthen Spike
			[51485]  = true, -- Earthgrab Totem
			[196884] = true, -- Feral Lunge
			[51533]  = true, -- Feral Spirit
			[193796] = true, -- Flametongue
			[196834] = true, -- Frostbrand
			[51514]  = true, -- Frog Hex
			[210873] = true, -- Raptor Hex
			[211004] = true, -- Spider Hex
			[211010] = true, -- Snake Hex
			[211015] = true, -- Roach Hex
			[192058] = true, -- Lightning Surge Totem
			[215864] = true, -- Rainfall
			[58875]  = true, -- Spirit Walk
			[201846] = true, -- Stormbringer
			[195222] = true, -- Stormlash
			[17364]  = true, -- Stormstrike
			[197214] = true, -- Sundering
			[196932] = true, -- Voodoo Totem
			[192077] = true, -- Wind Rush Totem
			[57994]  = true, -- Wind Shear
			[201898] = true, -- Windsong
		},
		[3] = { -- Restoration
			[108281] = true, -- Ancestral Guidance
			[207399] = true, -- Ancestral Protection Totem
			[114052] = true, -- Ascendance
			[108271] = true, -- Astral Shift
			--[2825]   = true, -- Bloodlust
			[157153] = true, -- Cloudburst Totem
			[198838] = true, -- Earthen Shield Totem
			[51485]  = true, -- Earthgrab Totem
			[188838] = true, -- Flame Shock
			[207778] = true, -- Gift of the Queen
			[192063] = true, -- Gust of Wind
			[73920]  = true, -- Healing Rain
			[5394]   = true, -- Healing Stream Totem
			[108280] = true, -- Healing Tide Totem
			[51514]  = true, -- Frog Hex
			[210873] = true, -- Raptor Hex
			[211004] = true, -- Spider Hex
			[211010] = true, -- Snake Hex
			[211015] = true, -- Roach Hex
			[51505]  = true, -- Lava Burst
			[77762]  = true, -- Lava Surge
			[192058] = true, -- Lightning Surge Totem
			[77130]  = true, -- Purify Spirit
			[61295]  = true, -- Riptide
			[98008]  = true, -- Spirit Link Totem
			[79206]  = true, -- Spiritwalker's Grace
			[53390]  = true, -- Tidal Waves
			[73685]  = true, -- Unleash Life
			[196932] = true, -- Voodoo Totem
			[197995] = true, -- Wellspring
			[192077] = true, -- Wind Rush Totem
			[57994]  = true, -- Wind Shear
		}
	}

	-- Abilities that require target check
	targetObj = {
		[1] = {
			--[51886]  = true, -- Cleanse Spirit
			[117014] = true, -- Elemental Blast
			[8042]   = true, -- Earth Shock
			[188389] = true, -- Flame Shock
			--[196840] = true, -- Frost Shock
			--[51514]  = true, -- Frog Hex
			[210714] = true, -- Icefury
			--[210873] = true, -- Raptor Hex
			--[211004] = true, -- Spider Hex
			--[211010] = true, -- Snake Hex
			--[211015] = true, -- Roach Hex
			[51505]  = true, -- Lava Burst
			--[57994]  = true, -- Wind Shear
		},
		[2] = {
			[201897] = true, -- Boulderfist
			--[51886]  = true, -- Cleanse Spirit
			[187874] = true, -- Crash Lightning
			[188089] = true, -- Earthen Spike
			[196884] = true, -- Feral Lunge
			[193796] = true, -- Flametongue
			[196834] = true, -- Frostbrand
			--[51514]  = true, -- Frog Hex
			--[210873] = true, -- Raptor Hex
			--[211004] = true, -- Spider Hex
			--[211010] = true, -- Snake Hex
			--[211015] = true, -- Roach Hex
			[17364]  = true, -- Stormstrike
			[197214] = true, -- Sundering
			[201898] = true, -- Windsong
		},
		[3] = {
			--[188838] = true, -- Flame Shock (Restoration)
			--[51514]  = true, -- Frog Hex
			--[210873] = true, -- Raptor Hex
			--[211004] = true, -- Spider Hex
			--[211010] = true, -- Snake Hex
			--[211015] = true, -- Roach Hex
			--[51505]  = true, -- Lava Burst
			--[77130]  = true, -- Purify Spirit
			--[61295]  = true, -- Riptide
			--[73685]  = true, -- Unleash Life
			--[57994]  = true, -- Wind Shear
		},
	}
	
	barObjIndex = {
		[114050] = "SPELL_AURA_APPLIED",	-- Ascendance (Elemental)
		[108281] = "SPELL_AURA_APPLIED",	-- Ancestral Guidance
		[108271] = "SPELL_AURA_APPLIED",	-- Astral Shift
		[118522] = "SPELL_AURA_APPLIED",	-- Elemental Blast (Critical Strike)
		[173183] = "SPELL_AURA_APPLIED",	-- Elemental Blast (Haste)
		[173184] = "SPELL_AURA_APPLIED",	-- Elemental Blast (Mastery)
		[16166]  = "SPELL_AURA_APPLIED",	-- Elemental Mastery
		[114051] = "SPELL_AURA_APPLIED",	-- Ascendance (Enhancement)
		[108271] = "SPELL_AURA_APPLIED",	-- Astral Shift
		[58875]  = "SPELL_AURA_APPLIED",	-- Spirit Walk
		[218825] = "SPELL_AURA_APPLIED",	-- Boulderfist
		[187878] = "SPELL_AURA_APPLIED",	-- Crash Lightning
		[193796] = "SPELL_CAST_SUCCESS",	-- Flametongue
		[196834] = "SPELL_AURA_APPLIED",	-- Frostbrand
		[201898] = "SPELL_AURA_APPLIED",	-- Windsong
		[188089] = "SPELL_CAST_SUCCESS",	-- Earthen Spike
		[51514]  = "SPELL_CAST_SUCCESS",	-- Hex
		[114052] = "SPELL_AURA_APPLIED",	-- Ascendance (Restoration)
		[108281] = "SPELL_AURA_APPLIED",	-- Ancestral Guidance
		[108271] = "SPELL_AURA_APPLIED",	-- Astral Shift
		[79206]  = "SPELL_AURA_APPLIED",	-- Spiritwalker's Grace
		[73685]  = "SPELL_AURA_APPLIED",	-- Unleash Life
	}
	
	barObj = {
		[1] = {
			["buff"] = {
				[114050] = SSA.AscendanceBarEle,
				[108281] = SSA.GuidanceBarEle,
				[108271] = SSA.AstralShiftBarEle,
				--[2825]   = SSA.BloodlustBarEle,
				[118522] = SSA.ElementalBlastCritBar,
				[173183] = SSA.ElementalBlastHasteBar,
				[173184] = SSA.ElementalBlastMasteryBar,
				[16166]  = SSA.ElementalMasteryBar,
			},
			["main"] = {
				[188616] = SSA.EarthElementalBar,
				[188592] = SSA.FireElementalBar,
				[157299] = SSA.StormElementalBar,
			},
			["utility"] = {
				[51485]  = SSA.EarthgrabTotemBarEle,
				[51514]  = SSA.HexBarEle,
				[192222] = SSA.LiquidMagmaTotemBar,
				[196932] = SSA.VoodooTotemBarEle,
				[192077] = SSA.WindRushTotemBarEle,
			},			
		},
		[2] = {
			["buff"] = {
				[114051] = SSA.AscendanceBarEnh,
				[108271] = SSA.AstralShiftBarEnh,
				--[2825]   = SSA.BloodlustBarEnh,
				[58875] = SSA.SpiritWalkBar,
			},
			["main"] = {
				[218825] = SSA.BoulderfistBar,
				[187878] = SSA.CrashLightningBar,
				[193796] = SSA.FlametongueBar,
				[196834] = SSA.FrostbrandBar,
				[201898] = SSA.WindsongBar,
			},
			["utility"] = {
				[188089] = SSA.EarthenSpikeBar,
				[51485]  = SSA.EarthgrabTotemBarEnh,
				[51514]  = SSA.HexBarEnh,
				[196932] = SSA.VoodooTotemBarEnh,
				[192077] = SSA.WindRushTotemBarEnh,
			},
		},
		[3] = {
			["buff"] = {
				[114052] = SSA.AscendanceBarRes,
				[108271] = SSA.AstralShiftBarRes,
				--[2825]   = SSA.BloodlustBarRes,
				[108281] = SSA.AncestralGuidanceBarRes,
				[79206]  = SSA.SpiritwalkersGraceBar,
				[73685]  = SSA.UnleashLifeBar,
			},
			["main"] = {
				[207399] = SSA.AncestralProtTotemBar,
				[157153] = SSA.CloudburstTotemBar,
				[5394]   = SSA.HealingStreamTotemBar,
				[108280] = SSA.HealingTideTotemBar,
				[98008]  = SSA.SpiritLinkTotemBar,
			},
			["utility"] = {
				[51485]  = SSA.EarthgrabTotemBarRes,
				[51514]  = SSA.HexBarRes,
				[196932] = SSA.VoodooTotemBarRes,
				[192077] = SSA.WindRushTotemBarRes,
			}
		}
	}
	
	
	miscObj = {
		[1] = {
			[1] = SSA.MaelstromBarEle,
			[3] = SSA.MainTimerBarGrpEle,
			[4] = SSA.UtilTimerBarGrpEle,
			[5] = SSA.BuffTimerBarGrpEle,
			[6] = SSA.TotemMasteryTexture,
		},
		[2] = {
			[1] = SSA.MaelstromBarEnh,
			[2] = SSA.MainTimerBarGrpEnh,
			[3] = SSA.UtilTimerBarGrpEnh,
			[4] = SSA.BuffTimerBarGrpEnh,
			[5] = SSA.StormstrikeChargeGrp,
		},
		[3] = {
			[1] = SSA.TidalWavesBar,
			[2] = SSA.MainTimerBarGrpRes,
			[3] = SSA.UtilTimerBarGrpRes,
			[4] = SSA.BuffTimerBarGrpRes,
		}
	}
	
	-- Assign Objects by Spell ID
	spellObj = {
		[1] = {
		-- Elemental
			[108281] = SSA.AncestralGuidanceEle,
			[114050] = SSA.AscendanceEle,
			[108271] = SSA.AstralShiftEle,
			--[2825]   = SSA.BloodlustEle,
			[51886]  = SSA.CleanseSpiritEle,
			[117014] = SSA.ElementalBlast,
			[198103] = SSA.EarthElemental,
			[51485]  = SSA.EarthgrabTotemEle,
			[8042]   = SSA.EarthShock,
			[16166]  = SSA.ElementalMastery,
			[198067] = SSA.FireElemental,
			[188389] = SSA.FlameShock,
			[192063] = SSA.GustWindEle,
			[51514]  = SSA.HexEle, -- Frog Hex
			[210873] = SSA.HexEle, -- Raptor Hex
			[211004] = SSA.HexEle, -- Spider Hex
			[211010] = SSA.HexEle, -- Snake Hex
			[211015] = SSA.HexEle, -- Roach Hex
			[210714] = SSA.Icefury,
			[51505]  = SSA.LavaBurstEle,
			[192058] = SSA.LightningSurgeTotemEle,
			[192222] = SSA.LiquidMagma,
			[192249] = SSA.StormElemental,
			[205495] = SSA.Stormkeeper,
			[51490]  = SSA.Thunderstorm,
			[196932] = SSA.VoodooTotemEle,
			[192077] = SSA.WindRushTotemEle,
			[57994]  = SSA.WindShearEle,
		},
		[2] = {		
			-- Enhancement
			[114051] = SSA.AscendanceEnh,
			[108271] = SSA.AstralShiftEnh,
			--[2825]   = SSA.BloodlustEnh,
			[201897] = SSA.Boulderfist,
			[51886]  = SSA.CleanseSpiritEnh,
			[187874] = SSA.CrashLightning,
			[188089] = SSA.EarthenSpike,
			[51485]  = SSA.EarthgrabTotemEnh,
			[196884] = SSA.FeralLunge,
			[51533]  = SSA.FeralSpirit,
			[193796] = SSA.Flametongue,
			[196834] = SSA.Frostbrand,
			[51514]  = SSA.HexEnh,
			[192058] = SSA.LightningSurgeTotemEnh,
			[215864] = SSA.Rainfall,
			[58875]  = SSA.SpiritWalk,
			[17364]  = SSA.Stormstrike,
			[197214] = SSA.Sundering,
			[196932] = SSA.VoodooTotemEnh,
			[192077] = SSA.WindRushTotemEnh,
			[201898] = SSA.Windsong,
			[57994]  = SSA.WindShearEnh,
		},
		[3] = {
			-- Restoration
			[108281] = SSA.AncestralGuidanceRes,
			[207399] = SSA.AncestralProtectionTotem,
			[114052] = SSA.AscendanceRes,
			[108271] = SSA.AstralShiftRes,
			--[2825]   = SSA.BloodlustRes,
			[157153] = SSA.CloudburstTotem,
			[198838] = SSA.EarthenShieldTotem,
			[51485]  = SSA.EarthgrabTotemRes,
			[188838] = SSA.FlameShockRes,
			[207778] = SSA.GiftOfQueen,
			[192063] = SSA.GustWindRes,
			[73920]  = SSA.HealingRain,
			[5394]   = SSA.HealingStreamTotem,
			[108280] = SSA.HealingTideTotem,
			[51514]  = SSA.HexRes,
			[51505]  = SSA.LavaBurstRes,
			--[51505]  = SSA.LavaBurst,
			[192058] = SSA.LightningSurgeTotemRes,
			[77130]  = SSA.PurifySpirit,
			[61295]  = SSA.Riptide,
			[98008]  = SSA.SpiritLinkTotem,
			[79206]  = SSA.SpiritwalkersGrace,
			[53390]  = SSA.TidalWaves,
			[73685]  = SSA.UnleashLife,
			[196932] = SSA.VoodooTotemRes,
			[197995] = SSA.Wellspring,
			[192077] = SSA.WindRushTotemRes,
			[57994]  = SSA.WindShearRes,
		}
	}

	-- Abilities with charges
	SSA.chargedSpells = {
		[1] = {
			[51505] = true,
		},
		[2] = {
			[201897] = true,
			[17364] = true,
		},
		[3] = {
			[51505] = true,
			[61295] = true,
			[5394] = true,
		}
	}
	
	Auras:UpdateTalents();
	spec = SSA.spec;
	
	local MaelstromBar = GetMaelstromBarObj(SSA.spec);
	local playerPower,maxPower = GetPowerInfo(L['Maelstrom']);
	local icefuryChgs = GetPowerInfo(L['Icefury']);
	
	InitializeStatusBar(MaelstromBar,0,maxPower,playerPower,0); -- Initialize Maelstrom Status Bar (Elemental)
	InitializeStatusBar(SSA.IcefuryBar,0,4,icefuryChgs); -- Initialize Icefury Status Bar
	InitializeStatusBar(SSA.IcefuryBar.Timer,0,15); -- Initialize Icefury Timer Status Bar
	InitializeStatusBar(SSA.EarthenShieldTotemBar.Timer,0,15) -- Initialize Earthen Shield Status Bar
	InitializeStatusBar(SSA.TidalWavesBar,0,2,0); -- Initialize Tidal Wave Status Bar
	
	-- Run Rotate Animation for Vertical Progress Bars
	for i=1,3 do
		for spellID in pairs(barObj[i].buff) do
			barObj[i].buff[spellID].rotatetime:Play();
			barObj[i].buff[spellID].rotatename:Play();
		end
		for spellID in pairs(barObj[i].main) do
			barObj[i].main[spellID].rotatetime:Play();
			barObj[i].main[spellID].rotatename:Play();
		end
		for spellID in pairs(barObj[i].utility) do
			barObj[i].utility[spellID].rotatetime:Play();
			barObj[i].utility[spellID].rotatename:Play();
		end
	end
	
	local dbEle = Auras.db.char.frames.eleGrp;
	for groupObj in pairs(dbEle) do
		InitializeGroup(SSA[groupObj],dbEle[groupObj].width,dbEle[groupObj].height,dbEle[groupObj].point,dbEle[groupObj].relativeTo,dbEle[groupObj].relativePoint,dbEle[groupObj].x,dbEle[groupObj].y,1);
	end
	
	--[[local dbEnh = Auras.db.char.frames.enhGrp;
	for groupObj in pairs(dbEnh) do
		InitializeGroup(SSA[groupObj],dbEnh[groupObj].width,dbEnh[groupObj].height,dbEnh[groupObj].point,dbEnh[groupObj].relativeTo,dbEnh[groupObj].relativePoint,dbEnh[groupObj].x,dbEnh[groupObj].y,2);
	end]]
	
	local dbRes = Auras.db.char.frames.resGrp;
	for groupObj in pairs(dbRes) do
		InitializeGroup(SSA[groupObj],dbRes[groupObj].width,dbRes[groupObj].height,dbRes[groupObj].point,dbRes[groupObj].relativeTo,dbRes[groupObj].relativePoint,dbRes[groupObj].x,dbRes[groupObj].y,3);
	end

	-- Continuously check for targets, their types, and if they're in range
	Frame:SetScript("OnUpdate",function(self)
		local spec = SSA.spec;
		local name = GetSpecName(spec):lower()
		local isExist = UnitExists('target');
		local isCombat = UnitAffectingCombat('player');
		local isFriend = UnitIsFriend('target','player');
		local isDead = UnitIsDead('target');
		
		if (spec == 3) then
			for spellID in pairs(spellObj[spec]) do
				if (isCombat or isFriend) then
					AdjustIconDisplay(spellObj[spec][spellID],1,1,1,1);
					if (Auras.db.char.aura[3].TidalWavesBar) then
						--SSA.TidalWavesBar.Flash:Stop();
						SSA.TidalWavesBar:Show();
					end
				else
					AdjustIconDisplay(spellObj[spec][spellID],0.5,0.5,0.5,Auras.db.char.triggers.res.OoCAlpha);
					if (Auras.db.char.aura[3].TidalWavesBar) then
						--SSA.TidalWavesBar.Flash:Stop();
						SSA.TidalWavesBar:Hide();
					end
				end
			end
		else
			if (isExist and isCombat) then
				if (not isFriend and not isDead) then
					if (TotemMasteryCheck()) then
						SSA.TotemMastery.Flash:Play();
						SSA.TotemMastery:Show();
						SSA.TotemMastery.Flash.fadeOut:SetToAlpha(1);
						SSA.TotemMastery.Flash.fadeIn:SetFromAlpha(1);
					else
						--SSA.TotemMastery.Flash:Stop();
						SSA.TotemMastery:Hide();
						--SSA.TotemMastery.Flash.fadeOut:SetToAlpha(1);
						--SSA.TotemMastery.Flash.fadeIn:SetFromAlpha(1);
					end
					for spellID in pairs(spellObj[spec]) do
						AdjustIconDisplay(spellObj[spec][spellID],1,1,1,1);
					end
				end
				for spellID in pairs(targetObj[spec]) do
					isRange = IsSpellInRange(GetSpellName(spellID),'target');
					
					if (not isFriend and not isDead) then
						if (isRange == 1) then
							AdjustIconDisplay(spellObj[spec][spellID],1,1,1,1);
						else
							AdjustIconDisplay(spellObj[spec][spellID],Auras.db.char.triggers[name].OoRColor.r,Auras.db.char.triggers[name].OoRColor.g,Auras.db.char.triggers[name].OoRColor.b,Auras.db.char.triggers[name].OoRColor.a);
						end
					elseif (isFriend) then
						AdjustIconDisplay(spellObj[spec][spellID],0.5,0.5,0.5,0.5);
					else
						AdjustIconDisplay(spellObj[spec][spellID],1,1,1,1);
					end
				end
			elseif (isExist and not isFriend and not isDead) then
				if (TotemMasteryCheck()) then
					SSA.TotemMastery.Flash:Play();
					SSA.TotemMastery:Show();
					SSA.TotemMastery.Flash.fadeOut:SetToAlpha(0.6);
					SSA.TotemMastery.Flash.fadeIn:SetFromAlpha(0.6);
				else
					--SSA.TotemMastery.Flash:Stop();
					SSA.TotemMastery:Hide();
				end
				for spellID in pairs(targetObj[spec]) do
					isRange = IsSpellInRange(GetSpellName(spellID),'target');
					
					if (not isFriend and not isDead) then
						if (isRange == 1) then
							AdjustIconDisplay(spellObj[spec][spellID],1,1,1,Auras.db.char.triggers[name].OoCAlpha);
						else
							AdjustIconDisplay(spellObj[spec][spellID],Auras.db.char.triggers[name].OoRColor.r,Auras.db.char.triggers[name].OoRColor.g,Auras.db.char.triggers[name].OoRColor.b,Auras.db.char.triggers[name].OoCAlpha);
						end
					else
						AdjustIconDisplay(spellObj[spec][spellID],1,1,1,Auras.db.char.triggers[name].OoCAlpha);
					end
				end
			else
				if (TotemMasteryCheck()) then
					--SSA.TotemMastery.Flash:Stop();
					SSA.TotemMastery:Hide();
				end
				for spellID in pairs(spellObj[spec]) do
					AdjustIconDisplay(spellObj[spec][spellID],1,1,1,Auras.db.char.triggers[name].OoCAlpha);
				end
			end
		end
	end);
end

local function ToggleOverlayGlow(object,toggle,enemyCheck)
	local isFriend = UnitIsFriend('target','player');
	if (enemyCheck and not isFriend) then
		if (toggle) then
			LibStub("LibButtonGlow-1.0").ShowOverlayGlow(object);
		else
			LibStub("LibButtonGlow-1.0").HideOverlayGlow(object);
		end
	else
		if (toggle) then
			LibStub("LibButtonGlow-1.0").ShowOverlayGlow(object);
		else
			LibStub("LibButtonGlow-1.0").HideOverlayGlow(object);
		end
	end
end

local function UpdateFlameShock(object,frame,spellID)
	--local name = GetSpellInfo(spellID);
	local debuff,_,_,_,_,duration,expires,caster = UnitDebuff('target',GetSpellName(spellID));
	
	
	if (not frame.text) then
		local fontsize = nil;
		if (spellID == 188838) then
			fontsize = 16;
		else
			fontsize = 18;
		end

		frame.text = frame:CreateFontString(nil, "HIGH", "GameFontHighlightLarge");
		frame.text:SetPoint("CENTER",0.5,0.5);
		frame.text:SetFont("Interface\\addons\\ShamanAuras\\media\\fonts\\PT_Sans_Narrow.TTF", fontsize,"OUTLINE");
		frame.text:SetTextColor(1,1,0,1);
	end
	
	object:SetScript("OnUpdate",nil);
	if (debuff and caster == 'player') then
		object:SetScript("OnUpdate", function(self,elapsed)
			local timer,seconds = parseTime(expires - GetTime(),false);

			frame.text:SetText(timer);
			
			if (seconds <= Auras.db.char.triggers.ele.flameShock) then
				--print("Timer: "..timer);
				ToggleOverlayGlow(object,true,true);
			else
				--print("Timerx: "..timer);
				ToggleOverlayGlow(object,false);
			end
			if (seconds == 0) then
				frame.text:SetText('');
				self:SetScript("OnUpdate",nil);
			end
		end);
	else
		frame.text:SetText('');
		if (UnitExists('target') and not UnitIsFriend('target','player') and not UnitIsDead('target') and UnitAffectingCombat('player')) then
			ToggleOverlayGlow(object,true,true);
		else
			ToggleOverlayGlow(object,false);
		end
	end
end

local function UpdateEarthenShield()
	local progress = ((Auras.db.char.info.totems.eShield.dmg or 0) / Auras.db.char.info.totems.eShield.hp * 100);
	progress = 100 - progress;
	local remains = (Auras.db.char.info.totems.eShield.hp - (Auras.db.char.info.totems.eShield.dmg or 0));
	
	if (remains > 0) then
		SSA.EarthenShieldTotemBar:SetValue(remains);
		SSA.EarthenShieldTotemBar.text:SetText(tostring(math.ceil(progress)).."%");
	else
		SSA.EarthenShieldTotemBar:Hide();
		SSA.EarthenShieldTotemBar.Timer:Hide();
	end
end

local function UpdateTidalWaveBar(alpha,color,isPlay,value)
	SSA.TidalWavesBar:SetStatusBarColor(color[1],color[2],color[3]);
	SSA.TidalWavesBar:SetValue(value);
	
	if (isPlay) then
		SSA.TidalWavesBar.Flash:Play();
	else
		SSA.TidalWavesBar.Flash:Stop();
	end
end

local function UpdateLavaSurge(object,spellID)
	--local name = GetSpellInfo(spellID);
	local buff,_,_,_,_,duration,expires = UnitBuff('player',GetSpellName(spellID));

	--[[object:SetScript("OnUpdate",nil);
	object:SetScript("OnUpdate", function(self,elapsed)
		if (buff) then
			local remaining = expires - GetTime();

			if (remaining > 0) then
				self:SetValue(remaining);
			else
				ToggleOverlayGlow(SSA.LavaBurstGlow,false);
				self:Hide();
				self.scaler:Stop();
				self:SetScript("OnUpdate",nil);
			end
		else
			self.scaler:Stop();
		end
	end);]]
end

--[[ To Enable Enhancement:
Auras:SetupCharges() -- SA_Core.lua
Auras:BuildMoveUI() -- SA_Init.lua
Auras:CreateVerticalStatusBar() -- SA_Init.lua
Auras:ChargeCooldown() -- SA.Core.lua
SA_Enhancement.lua
Auras:UNIT_POWER -- SA_Core.lua
SA_Core.lua (2255)
]]
function Auras:SetupCharges()
	local spec = SSA.spec;
	for spellID in pairs(SSA.chargedSpells[spec]) do
		local charges,maxCharges = GetSpellCharges(spellID);
		local object = spellObj[spec][spellID];

		--if ((SSA.isEoE or spec == 2) and spec ~= 2) then -- CHANGE FOR ENH
		if (SSA.isEoE and spec ~= 2) then
			if (spellID == 17364) then
				if (Auras.db.char.aura[2].StormstrikeChargesText and select(4,GetTalentInfo(5,1,1))) then
					object.Charges:Show();
					object.Charges.text:SetText("2");
				else
					object.Charges:Hide();
				end
			else
				object.Charges:Show();
				object.Charges.text:SetText("2");
			end
		else
			if (spec ~= 2) then
				object.Charges:Hide();
			end
		end
	end
end

function Auras:ChargeCooldown()
	if (SSA.spec ~= 2) then
		--local curText = SSA.ErrorFrame.text:GetText() or '';
		if (SSA.isEoE or select(4,GetTalentInfo(1,3,1)) or (select(4,GetTalentInfo(5,1,1)) and Auras.db.char.aura[2].StormstrikeChargesText)) then -- REMOVE "and spec ~=2"
			for spellID in pairs(SSA.chargedSpells[SSA.spec]) do
				local charges,maxCharges,start,duration = GetSpellCharges(spellID);
				local object = spellObj[SSA.spec][spellID];
				local ChargeCD = SSA[object:GetName()].ChargeCD;
				local ChargeText = SSA[object:GetName()].Charges;

				if (charges ~= maxCharges) then
					ChargeCD:SetCooldown(start,duration);
					ChargeText.text:SetText(charges);
				else
					ChargeText.text:SetText(charges);
				end
			end
		end
	end
end

function Auras:PLAYER_ENTERING_WORLD()
	--local name = GetSpellInfo(53390);
	local buff,_,_,count = UnitBuff('player',GetSpellName(53390));
	if (not buff) then
		UpdateTidalWaveBar(0,{1,0,0},true,2);
	else
		UpdateTidalWaveBar(0.5,{0.35,0.76,1},false,count);
	end
end

function Auras:ADDON_LOADED(event,addon)
	if (addon == "ShamanAuras") then
		
	end
	self:UnregisterEvent(event);
end

function Auras:PLAYER_LOGIN(event)
	ExecuteCooldown();
	
	self:UnregisterEvent(event);
end

function Auras:PLAYER_LEVEL_UP(event,level)
	LevelCheck(level);
end

function Auras:PLAYER_TALENT_UPDATE()
	Auras:UpdateTalents();
	local MaelstromBar = GetMaelstromBarObj(SSA.spec);
	local playerPower,maxPower = GetPowerInfo(L['Maelstrom']);
	InitializeStatusBar(MaelstromBar,0,maxPower,playerPower,0);
	spec = SSA.spec;
end

function Auras:PLAYER_TARGET_CHANGED()
	local spec = SSA.spec;
	--if (spec == 1 or spec == 2) then
	if (spec == 1) then
		if (spec == 1) then
			UpdateFlameShock(SSA.FlameShockGlow,SSA.FlameShock,188389);
		end
		MaelstromAlpha();
	elseif (spec == 3) then
		UpdateFlameShock(SSA.FlameShockResGlow,SSA.FlameShockRes,188838);
	end
	--[[if (isTotemMastery) then
		--local name = GetSpellInfo(210660);
		local buff = UnitBuff('player',GetSpellName(210660));
		if (buff or not IsEnemyTarget()) then 
			SSA.TotemMastery:Hide();
			SSA.TotemMasteryAlpha:Stop();
		else
			SSA.TotemMastery:Show();
			SSA.TotemMasteryAlpha:Play();
		end
	end]]
	--[[if (UnitExists('target')) then
		local fade = false;
		local isRange,isEnemy,isFriend,isDead = false,false,false,false;
		
		
	else
		Frame:SetScript("OnUpdate",nil);
	end]]
	--TargetCheck();
end

function Auras:PLAYER_REGEN_DISABLED()
	MaelstromAlpha();
end

function Auras:PLAYER_REGEN_ENABLED(event)
	MaelstromAlpha();
end

function Auras:UNIT_POWER(event,unit,powerType)
	local spec = SSA.spec;
	if (unit == 'player' and spec ~= 2 and Auras.db.char.triggers.ele.MaelstromAnim) then		
		if (powerType == "MAELSTROM" and spec ~= 3) then
			local MaelstromBar,trigger = GetMaelstromBarObj(spec);
			local playerPower,maxPower = GetPowerInfo("Maelstrom");

			if (playerPower >= trigger) then
				MaelstromBar.Lightning:SetAlpha(1);
				MaelstromBar.bg:SetAlpha(1);

				if (spec == 1) then
					ToggleOverlayGlow(SSA.EarthShockGlow,true);
				end
			else
				MaelstromBar.Lightning:SetAlpha(0);
				MaelstromBar.bg:SetAlpha(0.5);

				if (spec == 1) then
					ToggleOverlayGlow(SSA.EarthShockGlow,false);
				end
			end
			MaelstromBar:SetValue(playerPower);
			MaelstromBar.text:SetText(playerPower);
		end
	end
	
end
local function ToggleLightningOrbs(spellID)
	--local name = GetSpellInfo(spellID);
	local _,_,_,count = UnitBuff('player',GetSpellName(spellID));
	local elapsed = 0;

	if (count == 2) then
		StormstrikeCharge1:Show();
		StormstrikeCharge1.Lightning:SetScript("OnUpdate", function(self, elaps)
			elapsed = elapsed + (elaps * 1000);
			self:SetSequenceTime(1, elapsed);
		end);
		StormstrikeCharge2:Show();
		StormstrikeCharge2.Lightning:SetScript("OnUpdate", function(self, elaps)
			elapsed = elapsed + (elaps * 1000);
			self:SetSequenceTime(1, elapsed);
		end);
	elseif (count == 1) then
		
		if (select(4,GetTalentInfo(5,1,1))) then
			StormstrikeCharge2:Hide();
			StormstrikeCharge2.Lightning:SetScript("OnUpdate", nil);
		else
			StormstrikeCharge1:Show();
			StormstrikeCharge1.Lightning:SetScript("OnUpdate", function(self, elaps)
			elapsed = elapsed + (elaps * 1000);
			self:SetSequenceTime(1, elapsed);
		end);
		end
	elseif (count == 0) then
		StormstrikeCharge1:Hide();
		StormstrikeCharge1.Lightning:SetScript("OnUpdate", nil);
	end
end

function Auras:COMBAT_LOG_EVENT_UNFILTERED(eventname,...)
	if (eventname ~= "COMBAT_LOG_EVENT_UNFILTERED") then
		return;
	end
	
	local PlayerGUID = UnitGUID('player');
	local _,subevent,_,SourceGUID,source = ...;
	local spellID = select(12,...) or 0;
	local startTime = GetTime();
	
	if (SourceGUID == PlayerGUID or select(9,...) == "Earthen Shield Totem") then
		if (subevent == "SPELL_SUMMON" and spells[spec][spellID]) then
			if (spellID == 188592) then -- Fire Elemental
				isFireStormElemental = true;
				GUIDs[L["Greater Fire Elemental"]] = SSA.FireElementalBar;
				SSA.FireElementalBar:Show();
				SSA.FireElementalBar:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + 60;
					local timer,seconds = parseTime((expires - GetTime()),true);
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						isFireStormElemental = false;
						self:SetScript("OnUpdate",nil);
						self:Hide();
						GUIDs[L["Fire Elemental"]] = nil;
					end
				end);
			elseif (spellID == 157299) then -- Storm Elemental
				isFireStormElemental = true;
				GUIDs[L["Greater Storm Elemental"]] = SSA.StormElementalBar;
				SSA.StormElementalBar:Show();
				SSA.StormElementalBar:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + 60;
					local timer,seconds = parseTime((expires - GetTime()),true);
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						isFireStormElemental = false;
						self:SetScript("OnUpdate",nil);
						self:Hide();
						GUIDs[L["Storm Elemental"]] = nil;
					end
				end);
			elseif (spellID == 188616) then -- Earth Elemental
				isEarthElemental = true;
				GUIDs[L["Greater Earth Elemental"]] = SSA.EarthElementalBar;
				SSA.EarthElementalBar:Show();
				SSA.EarthElementalBar:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + 15;
					local timer,seconds = parseTime((expires - GetTime()),true);

					if (isFireStormElemental) then
						self:SetPoint("CENTER",230,-125);
					else
						self:SetPoint("CENTER",210,-125);
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						isEarthElemental = false;
						self:SetScript("OnUpdate",nil);
						self:Hide();
						GUIDs[L["Earth Elemental"]] = nil;
					end
				end);
			elseif (spellID == 198838) then -- Earthen Shield Totem
				local expires = GetTime() + 15;
				SSA.EarthenShieldTotemBar:SetMinMaxValues(0,UnitHealthMax('player')) -- Initialize Earthen Shield Status Bar
				SSA.EarthenShieldTotemBar:SetValue(UnitHealthMax('player'));
				SSA.EarthenShieldTotemBar:Show();
				SSA.EarthenShieldTotemBar.Timer:Show();
				Auras.db.char.info.totems.eShield.hp = UnitHealthMax('player');
				Auras.db.char.info.totems.eShield.dmg = 0;
				Auras.db.char.cooldowns[spellID] = GetTime();
				SSA.EarthenShieldTotemBar.Timer:SetScript("OnUpdate",function(self,elapsed)
					local timer,seconds = parseTime((expires - GetTime()),true);
					if (seconds > 0) then
						self:SetValue(seconds);
						self.text:SetText(timer);
					else
						self:SetValue(0);
						self.text:SetText('');
						SSA.EarthenShieldTotemBar:Hide();
						SSA.EarthenShieldTotemBar.Timer:Hide();
						self:SetScript("OnUpdate",nil);
					end
				end);
			elseif (spellID == 157153) then -- Cloudburst Totem
				Auras.db.char.info.totems.cBurst.absorb = 0;
				Auras.db.char.cooldowns[spellID] = GetTime();
				SSA.CloudburstCharges:Show();
				SSA.CloudburstCharges.text:SetText(0);
				SSA.CloudburstCharges:SetScript("OnUpdate",function(self,elapsed)
					local curTime = GetTime();
					--local name = GetSpellInfo(spellID);
					local buff,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,absorbed = UnitBuff("player", GetSpellName(spellID));
					local timeDiff = curTime - startTime;

					Auras.db.char.info.totems.cBurst.absorb = (absorbed or 0);
					self.text:SetText(Auras.db.char.info.totems.cBurst.absorb);
					if (not buff and timeDiff >= 2) then
						self.text:SetText('0');
						self:Hide();
						self:SetScript("OnUpdate",nil);
					end
				end);
			else
				Auras.db.char.cooldowns[spellID] = GetTime();
			end
			if (barObj[spec].main[spellID] and Auras.db.char.aura[spec][barObj[spec].main[spellID]:GetName()]) then
				local x = 0;
				
				table.insert(loadedMainBars,spellID);
				barObj[spec].main[spellID]:Show();
				barObj[spec].main[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);
				
					for i=1,getn(loadedMainBars) do
						if (loadedMainBars[i] == spellID) then
							self:SetPoint("RIGHT",((i-1) * -20) + -5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedMainBars) do
							if (loadedMainBars[i] == spellID) then
								table.remove(loadedMainBars,i);
							end
						end
					end
				end);
			end
			
			if (barObj[spec].utility[spellID] and Auras.db.char.aura[spec][barObj[spec].utility[spellID]:GetName()]) then
				local x = 0;
				
				table.insert(loadedUtilBars,spellID);
				barObj[spec].utility[spellID]:Show();
				barObj[spec].utility[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);

					x = (getn(loadedMainBars) * -25) + -190;
					for i=1,getn(loadedUtilBars) do
						if (loadedUtilBars[i] == spellID) then
							if (spec == 2 ) then
								SSA.UtilTimerBarGrpEnh:SetPoint("RIGHT",UIParent,"CENTER",x,-125);
							elseif (spec == 3) then
								SSA.UtilTimerBarGrpRes:SetPoint("RIGHT",UIParent,"CENTER",x,-125);
							end
							self:SetPoint("RIGHT",((i-1) * -20) + -5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedUtilBars) do
							if (loadedUtilBars[i] == spellID) then
								table.remove(loadedUtilBars,i);
							end
						end
					end
				end);
			end
		elseif (subevent == "SPELL_CAST_SUCCESS" and (spells[spec][spellID] or barObj[spec].buff[spellID] or barObj[spec].main[spellID])) then -- Lava Burst
			Auras.db.char.cooldowns[spellID] = GetTime();
			if (spellID == 196840 and isIcefury) then	-- Frost Shock & Icefury Check
				SSA.IcefuryBar:SetScript("OnUpdate", function(self,elapsed)
					local icefuryChgs = GetPowerInfo('Icefury');
					self:SetValue(icefuryChgs or 0);
				end);
			elseif (spellID == 193796) then -- Flametongue
				--[[spellObj[spec][spellID]:SetScript("OnUpdate", function(self,elapsed)
					--local name = GetSpellInfo(spellID);
					local _,_,_,_,_,_,expires = UnitBuff('player',GetSpellName(spellID));
					local remains = expires - GetTime();
					
					if (remains > 5) then
						ToggleOverlayGlow(SSA[self:GetName().."Glow"],false);
					elseif (remains <= 0.1) then
						self:SetScript("OnUpdate",nil);
					elseif (remains < 5) then
						ToggleOverlayGlow(SSA[self:GetName().."Glow"],true);
					end
				end);]]			
			elseif (spellID == 196834) then -- Frostbrand
				--[[spellObj[spec][spellID]:SetScript("OnUpdate", function(self,elapsed)
					--local name = GetSpellInfo(spellID);
					local _,_,_,_,_,_,expires = UnitBuff('player',GetSpellName(spellID));
					local remains = expires - GetTime();
					
					if (remains > 5) then
						ToggleOverlayGlow(SSA[self:GetName().."Glow"],false);
					elseif (remains <= 0.1) then
						self:SetScript("OnUpdate",nil);
					elseif (remains < 5) then
						ToggleOverlayGlow(SSA[self:GetName().."Glow"],true);
					end
				end);	]]		
			elseif (spellID == 201897) then -- Boulderfist
				--[[spellObj[spec][spellID]:SetScript("OnUpdate", function(self,elapsed)
					--local name = GetSpellInfo(spellID);
					local _,_,_,_,_,_,expires = UnitBuff('player',GetSpellName(spellID));
					local remains = expires - GetTime();
					
					if (remains > 5) then
						ToggleOverlayGlow(SSA[self:GetName().."Glow"],false);
					elseif (remains <= 0.1) then
						self:SetScript("OnUpdate",nil);
					elseif (remains < 5) then
						ToggleOverlayGlow(SSA[self:GetName().."Glow"],true);
					end
				end);	]]		
			end

			if (barObj[spec].buff[spellID] and Auras.db.char.aura[spec][barObj[spec].buff[spellID]:GetName()] and barObjIndex[spellID] == subevent) then
				local x = 0;
				
				table.insert(loadedBuffBars,spellID);
				barObj[SSA.spec].buff[spellID]:Show();
				barObj[SSA.spec].buff[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);
				
					for i=1,getn(loadedBuffBars) do
						if (loadedBuffBars[i] == spellID) then
							self:SetPoint("LEFT",((i-1) * 20) + 5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedBuffBars) do
							if (loadedBuffBars[i] == spellID) then
								table.remove(loadedBuffBars,i);
								--print("REMOVED: "..spellID);
							end
						end
					end
				end);
			elseif (barObj[spec].main[spellID] and Auras.db.char.aura[spec][barObj[spec].main[spellID]:GetName()] and barObjIndex[spellID] == subevent) then
				local x = 0;
				
				table.insert(loadedMainBars,spellID);
				barObj[spec].main[spellID]:Show();
				barObj[spec].main[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);
				
					for i=1,getn(loadedMainBars) do
						if (loadedMainBars[i] == spellID) then
							self:SetPoint("RIGHT",((i-1) * -20) + -5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedMainBars) do
							if (loadedMainBars[i] == spellID) then
								table.remove(loadedMainBars,i);
							end
						end
					end
				end);
			end
		elseif (subevent == "SPELL_AURA_APPLIED" and (spells[spec][spellID] or barObj[spec].buff[spellID] or barObj[spec].main[spellID])) then
			if (barObj[spec].buff[spellID] and Auras.db.char.aura[spec][barObj[spec].buff[spellID]:GetName()] and barObjIndex[spellID] == subevent) then
				local x = 0;
				table.insert(loadedBuffBars,spellID);
				barObj[SSA.spec].buff[spellID]:Show();
				barObj[SSA.spec].buff[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);

					for i=1,getn(loadedBuffBars) do
						if (loadedBuffBars[i] == spellID) then
							--SSA.BuffTimerBarGrpEle:SetPoint("CENTER",x,-125);
							self:SetPoint("LEFT",((i-1) * 20) + 5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedBuffBars) do
							if (loadedBuffBars[i] == spellID) then
								table.remove(loadedBuffBars,i);
								--print("REMOVED: "..spellID);
							end
						end
					end
				end);
			elseif (barObj[spec].main[spellID] and Auras.db.char.aura[spec][barObj[spec].main[spellID]:GetName()] and barObjIndex[spellID] == subevent) then
				local x = 0;
				
				table.insert(loadedMainBars,spellID);
				barObj[spec].main[spellID]:Show();
				barObj[spec].main[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);
				
					for i=1,getn(loadedMainBars) do
						if (loadedMainBars[i] == spellID) then
							self:SetPoint("RIGHT",((i-1) * -20) + -5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedMainBars) do
							if (loadedMainBars[i] == spellID) then
								table.remove(loadedMainBars,i);
							end
						end
					end
				end);
			elseif (spellID == 205495 and Auras.db.char.aura[1].StormkeeperChargeGrp) then -- Stormkeeper
				local buff,count
				SSA.StormkeeperChargeGrp:SetScript("OnUpdate",function(self,elapsed)
					buff,_,_,count = UnitBuff('player',GetSpellName(spellID))
					
					self:Show();
					if (count == 3) then
						self.Charge1:SetAlpha(1);
						self.Charge2:SetAlpha(1);
						self.Charge3:SetAlpha(1);
					elseif (count == 2) then
						--self.Charge1:Show();
						--self.Charge2:Show();
						self.Charge3:SetAlpha(0);
					elseif (count == 1) then
						--self.Charge1:Show();
						self.Charge2:SetAlpha(0);
						--self.Charge3:Hide();
					elseif (count == 0 or not buff) then
						self.Charge1:SetAlpha(0);
						self.Charge2:SetAlpha(0);
						self.Charge3:SetAlpha(0);
						self:SetScript("OnUpdate",nil);
					end
				end);
			elseif (spellID == 210714) then -- Icefury
				isIcefury = true;
				execTime[spellID] = GetTime();
				SSA.Icefury.CD.text:Show();
				SSA.IcefuryBar:Show();
				SSA.IcefuryBar:SetValue(4);
				SSA.IcefuryBar.Timer:SetScript("OnUpdate",function(self,elapsed)
					--local name = GetSpellInfo(spellID);
					local buff,_,_,_,_,duration,expires = UnitBuff('player',GetSpellName(spellID));
					local timer,seconds = parseTime((expires - GetTime()),true);

					if (seconds > 0) then
						self:SetValue(seconds);
						self.text:SetText(timer);
					else
						self:SetScript("OnUpdate",nil);
					end
				end);
			elseif (spellID == 77762) then -- Lava Surge
				local name
				if (SSA.spec == 3) then
					name = "Res";
				else
					name = 'Ele';
					ToggleOverlayGlow(SSA["LavaBurst"..name.."Glow"],true,false);
				end
				SSA["LavaSurge"..name]:Show();
				SSA["LavaSurge"..name].Scaler:Play();
				SSA["LavaBurst"..name].CD:Hide();
				--SSA["LavaBurst"..name].CD:SetSwipeColor(1,1,1,0);
				SSA["LavaBurst"..name].CD:SetCooldown(0,0)
				ToggleCooldownSwipe(SSA["LavaBurst"..name].CD,false,true);
				--UpdateLavaSurge(SSA["LavaSurge"..name],spellID);
				--[[if (CDFrame[spec][51505]) then
					CDFrame[spec][51505]:Hide();
					CDFrame[spec][51505]:SetCooldown(0,0);
				end]]
			elseif (spellID == 188389) then -- Flame Shock
				UpdateFlameShock(SSA.FlameShockGlow,spellObj[spec][spellID],spellID);
			elseif (spellID == 188838) then
				UpdateFlameShock(SSA.FlameShockResGlow,spellObj[spec][spellID],spellID);
			elseif (spellID == 53390) then -- Tidal Waves
				--local name = GetSpellInfo(spellID);
				local _,_,_,count = UnitBuff('player',GetSpellName(spellID));
							
				UpdateTidalWaveBar(0.5,{0.35,0.76,1},false,count);
			elseif (spellID == 201846 and Auras.db.char.aura[2].StormstrikeChargesOrbs and select(4,GetTalentInfo(5,1,1))) then
				local name,count = GetSpellInfo(spellID),nil;				
				local elapsed = 0;
				
				StormstrikeCharge1.Lightning:SetScript("OnUpdate", function(self, elaps)
					elapsed = elapsed + (elaps * 1000);
					self:SetSequenceTime(1, elapsed);
				end);
				
				StormstrikeCharge2.Lightning:SetScript("OnUpdate", function(self, elaps)
					elapsed = elapsed + (elaps * 1000);
					self:SetSequenceTime(1, elapsed);
				end);
				
				SSA.StormstrikeChargeGrp:SetScript("OnUpdate",function(self,elaps)
					count = select(4,UnitBuff('player',name));
					if (count == 2) then
						StormstrikeCharge1:Show();
						StormstrikeCharge2:Show();
					elseif (count == 1) then
						if (select(4,GetTalentInfo(5,1,1))) then
							StormstrikeCharge2:Hide();
						else
							StormstrikeCharge1:Show();
						end
					else
						StormstrikeCharge1:Hide();
						self:SetScript("OnUpdate",nil);
					end
				end);
			end
		elseif ((subevent == "SPELL_AURA_APPLIED_DOSE" or subevent == "SPELL_AURA_REMOVED_DOSE") and spells[spec][spellID]) then
			if (spellID == 53390) then -- Tidal Waves
				--print("TIDAL WAVES REFRESH!");
				--local name = GetSpellInfo(spellID);
				local _,_,_,count = UnitBuff('player',GetSpellName(spellID));
				SSA.TidalWavesBar:SetValue(count);
			end	
		elseif (subevent == "SPELL_AURA_REFRESH" and spells[spec][spellID]) then
			if (spellID == 77762) then
				local name
				if (SSA.spec == 3) then
					name = "Res";
				else
					name = 'Ele';
				end
				SSA["LavaBurst"..name].CD:Hide();
				SSA["LavaBurst"..name].CD:SetCooldown(0,0)
				ToggleCooldownSwipe(SSA["LavaBurst"..name].CD,false,true);
				--UpdateLavaSurge(SSA["LavaSurge"..name],spellID);
			elseif (spellID == 188389) then
				UpdateFlameShock(SSA.FlameShockGlow,SSA.FlameShock,spellID);
			elseif (spellID == 188838) then
				UpdateFlameShock(SSA.FlameShockResGlow,SSA.FlameShockRes,spellID);
			end
			if (barObj[spec].main[spellID] and Auras.db.char.aura[spec][barObj[spec].main[spellID]:GetName()] and barObjIndex[spellID] == "SPELL_AURA_APPLIED") then
				local x = 0;
				
				table.insert(loadedMainBars,spellID);
				barObj[spec].main[spellID]:Show();
				barObj[spec].main[spellID]:SetScript("OnUpdate", function(self,elapsed)
					local expires = startTime + self.duration;
					local timer,seconds = parseTime((expires - GetTime()),true);
				
					for i=1,getn(loadedMainBars) do
						if (loadedMainBars[i] == spellID) then
							self:SetPoint("RIGHT",((i-1) * -20) + -5,0);
						end
					end
					if (seconds > 0) then
						self.timetext:SetText(timer);
						self:SetValue(seconds);
					else
						self:SetScript("OnUpdate",nil);
						self:Hide();
						for i=1,getn(loadedMainBars) do
							if (loadedMainBars[i] == spellID) then
								table.remove(loadedMainBars,i);
							end
						end
					end
				end);
			end
		elseif (subevent == "SPELL_AURA_REMOVED" and spells[spec][spellID]) then
			if (spellID == 210714) then
				isIcefury = false;
				execTime[spellID] = nil;
				SSA.Icefury.CD:SetScript("OnUpdate",nil);
				SSA.IcefuryBar:SetScript("OnUpdate",nil);
				SSA.Icefury.CD:Hide();
				SSA.IcefuryBar:Hide();
			elseif (spellID == 77762) then
				local name
				if (SSA.spec == 3) then
					name = "Res";
				else
					name = 'Ele';
					ToggleOverlayGlow(SSA["LavaBurst"..name.."Glow"],false);
				end
				SSA["LavaSurge"..name]:Hide();
				SSA["LavaSurge"..name].Scaler:Stop();
				--SSA.LavaSurgeBar:Hide();
				--SSA.LavaSurgeBar.scaler:Stop();
				--ToggleOverlayGlow(SSA.LavaBurstGlow,false);
			elseif (spellID == 53390) then -- Tidal Waves
				UpdateTidalWaveBar(0,{1,0,0},true,2);
			elseif (spellID == 201846 and Auras.db.char.aura[2].StormstrikeChargesOrbs and select(4,GetTalentInfo(5,1,1))) then
				StormstrikeCharge1.Lightning:SetScript("OnUpdate", nil);
				StormstrikeCharge2.Lightning:SetScript("OnUpdate", nil);
			end
		elseif (subevent == "SPELL_DAMAGE" and spec == 3) then
			for i=1,5 do
				local _,name = GetTotemInfo(i);
				if (name == "Earthen Shield Totem") then
					Auras.db.char.info.totems.eShield.dmg = (Auras.db.char.info.totems.eShield.dmg or 0) + select(15,...);
					UpdateEarthenShield();
					break;
				end
			end
		elseif (subevent == "SWING_DAMAGE" and spec == 3) then
			for i=1,5 do
				local _,name = GetTotemInfo(i);
				if (name == "Earthen Shield Totem") then
					Auras.db.char.info.totems.eShield.dmg = (Auras.db.char.info.totems.eShield.dmg or 0) + (select(15,...) or 0);
					UpdateEarthenShield();
					break;
				end
			end
		elseif (subevent == "UNIT_DIED") then
			for elemental,barObj in pairs(GUIDs) do
				if (select(9,...) == elemental) then
					barObj:SetScript("OnUpdate",nil);
					barObj:Hide();
					GUIDs[elemental] = nil;
				end
			end
		end
	end
end

function Auras:SPELL_UPDATE_COOLDOWN()
	ExecuteCooldown();
end

